@@ -6,24 +6,37 @@ Documents
 from __future__ import division, print_function, unicode_literals
 
 # Standard Library
+import logging
 import os
 import re
 import warnings
 from functools import partial
 
 # Third Party
-import requests
 from future.utils import python_2_unicode_compatible
+from requests.exceptions import RequestException
 
 # Local
 from .annotations import AnnotationClient
 from .base import APIResults, BaseAPIClient, BaseAPIObject
 from .constants import BULK_LIMIT
+from .exceptions import APIError
 from .organizations import Organization
 from .sections import SectionClient
-from .toolbox import grouper, is_url, merge_dicts
+from .toolbox import grouper, is_url, merge_dicts, requests_retry_session
 from .users import User
 
+try:
+    # Standard Library
+    from urllib.parse import urlparse
+except ImportError:
+    # Third Party
+    from urlparse import urlparse
+
+
+
+logger = logging.getLogger("documentcloud")
+
 
 @python_2_unicode_compatible
 class Document(BaseAPIObject):
@@ -35,10 +48,12 @@ class Document(BaseAPIObject):
         "data",
         "description",
         "language",
-        "related_article",
+        "publish_at",
         "published_url",
+        "related_article",
         "source",
         "title",
+        "delayed_index",
     ]
     date_fields = ["created_at", "updated_at"]
 
@@ -115,7 +130,7 @@ class Document(BaseAPIObject):
 
     @property
     def mentions(self):
-        if hasattr(self, "highlights"):
+        if hasattr(self, "highlights") and self.highlights is not None:
             return [
                 Mention(page, text)
                 for page, texts in self.highlights.items()
@@ -151,9 +166,20 @@ class Document(BaseAPIObject):
         return self.organization.slug
 
     def _get_url(self, url, text):
-        response = requests.get(url, headers={"User-Agent": "python-documentcloud2"})
+        base_netloc = urlparse(self._client.base_uri).netloc
+        url_netloc = urlparse(url).netloc
+
+        if base_netloc == url_netloc:
+            # if the url host is the same as the base api host,
+            # sent the request with the client in order to include
+            # authentication credentials
+            response = self._client.get(url, full_url=True)
+        else:
+            response = requests_retry_session().get(
+                url, headers={"User-Agent": "python-documentcloud2"}
+            )
         if text:
-            return response.content.decode("utf8")
+            return response.text
         else:
             return response.content
 
@@ -182,6 +208,10 @@ class Document(BaseAPIObject):
             self.get_image_url(page=i, size=size) for i in range(1, self.page_count + 1)
         ]
 
+    def process(self):
+        """Reprocess the document"""
+        self._client.post("{}/{}/process/".format(self.api_path, self.id))
+
 
 class DocumentClient(BaseAPIClient):
     """Client for interacting with Documents"""
@@ -192,13 +222,9 @@ class DocumentClient(BaseAPIClient):
     def search(self, query, **params):
         """Return documents matching a search query"""
 
+        # legacy parameter mentions renamed to hl
         mentions = params.pop("mentions", None)
-        if mentions is not None:  # pragma: no cover
-            warnings.warn(
-                "The `mentions` argument to `search` is deprecated, "
-                "it will always include mentions from all pages now",
-                DeprecationWarning,
-            )
+        params["hl"] = mentions
         data = params.pop("data", None)
         if data is not None:  # pragma: no cover
             warnings.warn(
@@ -212,6 +238,12 @@ class DocumentClient(BaseAPIClient):
         response = self.client.get("documents/search/", params=params)
         return APIResults(self.resource, self.client, response)
 
+    def list(self, **params):
+        """Convert id__in from list to string if needed"""
+        if "id__in" in params and isinstance(params["id__in"], list):
+            params["id__in"] = ",".join(str(i) for i in params["id__in"])
+        return super().list(**params)
+
     def upload(self, pdf, **kwargs):
         """Upload a document"""
         # if they pass in a URL, use the URL upload flow
@@ -244,7 +276,9 @@ class DocumentClient(BaseAPIClient):
             "access",
             "description",
             "language",
+            "original_extension",
             "related_article",
+            "publish_at",
             "published_url",
             "source",
             "title",
@@ -294,7 +328,7 @@ class DocumentClient(BaseAPIClient):
         # upload the file directly to storage
         create_json = response.json()
         presigned_url = create_json["presigned_url"]
-        response = requests.put(presigned_url, data=file_.read())
+        response = requests_retry_session().put(presigned_url, data=file_.read())
 
         # begin processing the document
         doc_id = create_json["id"]
@@ -304,7 +338,7 @@ class DocumentClient(BaseAPIClient):
 
         return Document(self.client, create_json)
 
-    def _collect_files(self, path):
+    def _collect_files(self, path, extension):
         """Find the paths to all pdfs under a directory"""
         path_list = []
         for (dirpath, _dirname, filenames) in os.walk(path):
@@ -312,48 +346,94 @@ class DocumentClient(BaseAPIClient):
                 [
                     os.path.join(dirpath, i)
                     for i in filenames
-                    if i.lower().endswith(".pdf")
+                    if extension is None or i.lower().endswith(extension)
                 ]
             )
         return path_list
 
-    def upload_directory(self, path, **kwargs):
+    def upload_directory(self, path, handle_errors=False, extension=".pdf", **kwargs):
         """Upload all PDFs in a directory"""
 
         # do not set the same title for all documents
         kwargs.pop("title", None)
 
         # Loop through the path and get all the files
-        path_list = self._collect_files(path)
+        path_list = self._collect_files(path, extension)
+
+        logger.info(
+            "Upload directory on %s: Found %d files to upload", path, len(path_list)
+        )
 
         # Upload all the pdfs using the bulk API to reduce the number
         # of API calls and improve performance
         obj_list = []
         params = self._format_upload_parameters("", **kwargs)
-        for pdf_paths in grouper(path_list, BULK_LIMIT):
+        for i, pdf_paths in enumerate(grouper(path_list, BULK_LIMIT)):
             # Grouper will put None's on the end of the last group
             pdf_paths = [p for p in pdf_paths if p is not None]
 
+            logger.info("Uploading group %d: %s", i + 1, "\n".join(pdf_paths))
+
             # create the documents
-            response = self.client.post(
-                "documents/",
-                json=[
-                    merge_dicts(params, {"title": self._get_title(p)})
-                    for p in pdf_paths
-                ],
-            )
+            logger.info("Creating the documents...")
+            try:
+                response = self.client.post(
+                    "documents/",
+                    json=[
+                        merge_dicts(params, {"title": self._get_title(p)})
+                        for p in pdf_paths
+                    ],
+                )
+            except (APIError, RequestException) as exc:
+                if handle_errors:
+                    logger.info(
+                        "Error creating the following documents: %s %s",
+                        exc,
+                        "\n".join(pdf_paths),
+                    )
+                    continue
+                else:
+                    raise
 
             # upload the files directly to storage
             create_json = response.json()
             obj_list.extend(create_json)
             presigned_urls = [j["presigned_url"] for j in create_json]
             for url, pdf_path in zip(presigned_urls, pdf_paths):
-                response = requests.put(url, data=open(pdf_path, "rb").read())
-                self.client.raise_for_status(response)
+                logger.info("Uploading %s to S3...", pdf_path)
+                try:
+                    response = requests_retry_session().put(
+                        url, data=open(pdf_path, "rb").read()
+                    )
+                    self.client.raise_for_status(response)
+                except (APIError, RequestException) as exc:
+                    if handle_errors:
+                        logger.info(
+                            "Error uploading the following document: %s %s",
+                            exc,
+                            pdf_path,
+                        )
+                        continue
+                    else:
+                        raise
 
             # begin processing the documents
+            logger.info("Processing the documents...")
             doc_ids = [j["id"] for j in create_json]
-            response = self.client.post("documents/process/", json={"ids": doc_ids})
+            try:
+                response = self.client.post("documents/process/", json={"ids": doc_ids})
+            except (APIError, RequestException) as exc:
+                if handle_errors:
+                    logger.info(
+                        "Error creating the following documents: %s %s",
+                        exc,
+                        "\n".join(pdf_paths),
+                    )
+                    continue
+                else:
+                    raise
+
+        logger.info("Upload directory complete")
 
         # Pass back the list of documents
         return [Document(self.client, d) for d in obj_list]
